
#include <./libsetrecon/reconciled_set.h>

#ifdef __cplusplus
extern "C" {
#endif
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"
#include "ppport.h"
#ifdef __cplusplus
}
#endif

#include <sstream>

#include "const-c.inc"

MODULE = SetRecon		PACKAGE = SetRecon		

INCLUDE: const-xs.inc

SetRecon*
SetRecon::new(int b, int m)
	CODE:
		SetRecon::setField(b, m);
		SetSeed(NTL::to_ZZ(time(NULL)));
		RETVAL = new SetRecon(b, m);
		ST(0) = sv_newmortal();
		sv_setref_pv( ST(0), CLASS, (void*)RETVAL );

void
SetRecon::DESTROY()

Sketch* 
SetRecon::GetSketch()
	CODE:
		RETVAL = THIS->getPoly();
		ST(0) = sv_newmortal();
		sv_setref_pv( ST(0), "SetRecon", (void*)RETVAL );

void
SetRecon::SetSketchData(sketch)
		SV*	sketch
	CODE:
		AV* hashes = (AV*) SvRV(*hv_fetch((HV *)SvRV(sketch), "hashes", 6, 0));
		AV* extra_hashes = (AV*) SvRV(*hv_fetch((HV *)SvRV(sketch), "extra_hashes", 12, 0));
		AV* valids = (AV*) SvRV(*hv_fetch((HV *)SvRV(sketch), "valids", 6, 0));

		int len = av_len(hashes) + 1;
		THIS->getPoly()->hash_vector.SetLength(len);
		for (long i = 0; i < len; i++)
		{
			NTL::ZZ_p r = NTL::to_ZZ_p( NTL::to_ZZ(SvPV_nolen(*av_fetch(hashes, i, 0))) );
			THIS->getPoly()->hash_vector[i] = r; 
		}

		len = av_len(extra_hashes) + 1;
		THIS->getPoly()->extra_hashes.SetLength(len);
		for (long i = 0; i < len; i++)
		{
			NTL::ZZ_p r = NTL::to_ZZ_p( NTL::to_ZZ(SvPV_nolen(*av_fetch(extra_hashes, i, 0))) );
			THIS->getPoly()->extra_hashes[i] = r; 
		}
		
		len = av_len(valids) + 1;
		THIS->getPoly()->valids.SetLength(len);
		for (long i = 0; i < len; i++)
		{
			NTL::ZZ_p r = NTL::to_ZZ_p( NTL::to_ZZ(SvPV_nolen(*av_fetch(valids, i, 0))) );
			THIS->getPoly()->valids[i] = r; 
		}

		THIS->getPoly()->setsize = SvIV(*hv_fetch((HV *)SvRV(sketch), "size", 4, 0));

HV*
SetRecon::GetSketchData()
	INIT:
		HV* result;
		AV* nhashes, *extra_hashes, *valids;
	CODE:
		result = (HV *)sv_2mortal((SV *)newHV()); 
		nhashes = (AV *)sv_2mortal((SV *)newAV());
		extra_hashes = (AV *)sv_2mortal((SV *)newAV());
		valids = (AV *)sv_2mortal((SV *)newAV());

		NTL::vec_ZZ_p& hashes = THIS->getPoly()->hash_vector;

		for (long i = 0; i < hashes.length(); i++)
		{
    			std::ostringstream tmp; 
    			tmp << hashes[i];
    			av_push(nhashes, newSVpv(tmp.str().c_str(), tmp.str().length()));
		}

		NTL::vec_ZZ_p& rextra_hashes = THIS->getPoly()->extra_hashes;

		for (long i = 0; i < rextra_hashes.length(); i++)
		{
    			std::ostringstream tmp; 
    			tmp << rextra_hashes[i];
    			av_push(extra_hashes, newSVpv(tmp.str().c_str(), tmp.str().length()));
		}

		NTL::vec_ZZ_p& rvalids = THIS->getPoly()->valids;

		for (long i = 0; i < rvalids.length(); i++)
		{
    			std::ostringstream tmp; 
    			tmp << rvalids[i];
    			av_push(valids, newSVpv(tmp.str().c_str(), tmp.str().length()));
		}

		hv_store(result, "size", 4, newSViv(THIS->getPoly()->setsize), 0);
		hv_store(result, "hashes", 6, newRV((SV*)nhashes), 0);
		hv_store(result, "extra_hashes", 12, newRV((SV*)extra_hashes), 0);
		hv_store(result, "valids", 6, newRV((SV*)valids), 0);
		RETVAL = result;
	OUTPUT:
		RETVAL

void
SetRecon::AddElement(element)
		const char*	element
	CODE:
		NTL::ZZ_p r = NTL::to_ZZ_p( NTL::to_ZZ(element) );
		THIS->add_element(r); 

void
SetRecon::RemoveSketch()
	CODE:
		THIS->getPoly()->reset();

void
SetRecon::AddSet(set_str)
		SV*	set_str
	INIT:
		NTL::ZZ_p r;
		int len;

		if ((!SvROK(set_str))
			|| (SvTYPE(SvRV(set_str)) != SVt_PVAV)
			|| ((len = av_len((AV *)SvRV(set_str))) < 0))
		{
			XSRETURN_UNDEF;
		}
	CODE:
		for (long i = 0; i < len + 1; i++)
		{
			r = NTL::to_ZZ_p( NTL::to_ZZ(SvPV_nolen(*av_fetch((AV *)SvRV(set_str), i, 0))) );
			THIS->add_element(r); 
		}

SV*
SetRecon::GetDiff(p_other, out_set_A, out_set_B)
		HV*	p_other
		SV*	out_set_A
		SV*	out_set_B
	INIT:
		NTL::vec_ZZ_p diffA, diffB;

		AV * diffA_AV, * diffB_AV;

		diffA_AV = (AV*) SvRV(out_set_A);
		diffB_AV = (AV*) SvRV(out_set_B);
	CODE:
		AV* hashes = (AV*) SvRV(*hv_fetch(p_other, "hashes", 6, 0));
		AV* extra_hashes = (AV*) SvRV(*hv_fetch(p_other, "extra_hashes", 12, 0));
		AV* valids = (AV*) SvRV(*hv_fetch(p_other, "valids", 6, 0));

		int len = av_len(hashes) + 1;
		NTL::vec_ZZ_p tmp1;
		tmp1.SetLength(len);
		for (long i = 0; i < len; i++)
		{
			NTL::ZZ_p r = NTL::to_ZZ_p( NTL::to_ZZ(SvPV_nolen(*av_fetch(hashes, i, 0))) );
			tmp1[i] = r; 
		}

		len = av_len(extra_hashes) + 1;
		NTL::vec_ZZ_p tmp2;
		tmp2.SetLength(len);
		for (long i = 0; i < len; i++)
		{
			NTL::ZZ_p r = NTL::to_ZZ_p( NTL::to_ZZ(SvPV_nolen(*av_fetch(extra_hashes, i, 0))) );
			tmp2[i] = r; 
		}
		
		len = av_len(valids) + 1;
		NTL::vec_ZZ_p tmp3;
		tmp3.SetLength(len);
		for (long i = 0; i < len; i++)
		{
			NTL::ZZ_p r = NTL::to_ZZ_p( NTL::to_ZZ(SvPV_nolen(*av_fetch(valids, i, 0))) );
			tmp3[i] = r; 
		}

		Sketch sketch(tmp1, tmp2, tmp3, SvIV(*hv_fetch(p_other, "size", 4, 0)));

		int res = THIS->Reconcile(&sketch, diffA, diffB); 

		if (res > 0)
		{
			for (long i = 0; i < diffA.length(); i++)
			{
				std::ostringstream tmp; 
				tmp << diffA[i];
				av_push(diffA_AV, newSVpv(tmp.str().c_str(), tmp.str().length()));
			}

			for (long i = 0; i < diffB.length(); i++)
			{
				std::ostringstream tmp; 
				tmp << diffB[i];
				av_push(diffB_AV, newSVpv(tmp.str().c_str(), tmp.str().length()));
			}
		}
		RETVAL = newSViv(res);
	OUTPUT:
		RETVAL

SV*
SetRecon::GetDiffSizes(p_other)
		HV*	p_other
	INIT:
		AV* result;
		unsigned int diffASize, diffBSize;

	CODE:
		result = (AV *)sv_2mortal((SV *)newAV()); 
		AV* hashes = (AV*) SvRV(*hv_fetch(p_other, "hashes", 6, 0));
		AV* extra_hashes = (AV*) SvRV(*hv_fetch(p_other, "extra_hashes", 12, 0));
		AV* valids = (AV*) SvRV(*hv_fetch(p_other, "valids", 6, 0));

		int len = av_len(hashes) + 1;
		NTL::vec_ZZ_p tmp1;
		tmp1.SetLength(len);
		for (long i = 0; i < len; i++)
		{
			NTL::ZZ_p r = NTL::to_ZZ_p( NTL::to_ZZ(SvPV_nolen(*av_fetch(hashes, i, 0))) );
			tmp1[i] = r; 
		}

		len = av_len(extra_hashes) + 1;
		NTL::vec_ZZ_p tmp2;
		tmp2.SetLength(len);
		for (long i = 0; i < len; i++)
		{
			NTL::ZZ_p r = NTL::to_ZZ_p( NTL::to_ZZ(SvPV_nolen(*av_fetch(extra_hashes, i, 0))) );
			tmp2[i] = r; 
		}
		
		len = av_len(valids) + 1;
		NTL::vec_ZZ_p tmp3;
		tmp3.SetLength(len);
		for (long i = 0; i < len; i++)
		{
			NTL::ZZ_p r = NTL::to_ZZ_p( NTL::to_ZZ(SvPV_nolen(*av_fetch(valids, i, 0))) );
			tmp3[i] = r; 
		}

		Sketch sketch(tmp1, tmp2, tmp3, SvIV(*hv_fetch(p_other, "size", 4, 0)));

		int res = THIS->Reconcile(&sketch, diffASize, diffBSize); 

		av_push(result, newSViv( res > 0 ? (int) diffASize : -1));
		av_push(result, newSViv( res > 0 ? (int) diffBSize : -1));
		RETVAL = newRV((SV *)result);

	OUTPUT:
		RETVAL


