#ifndef POLYNOMIAL
#define POLYNOMIAL

// A Sketch in the partition tree
#include <NTL/matrix.h>
#include <NTL/mat_ZZ_p.h>
#include <NTL/vec_ZZ.h>
#include <NTL/ZZ.h>
#include <NTL/ZZ_p.h>
#include <NTL/vec_ZZ_p.h>
#include <math.h>
#include <NTL/vec_vec_ZZ_p.h>
#include <NTL/ZZ_pX.h>
#include <NTL/ZZ_pXFactoring.h>

using namespace std;
NTL_CLIENT

class Sketch
/* This class defines Sketchs in the partition-tree data structure used to amortize
 **   the computational cost of characteristic Sketch evaluations over insertions
 **   and deletions into the database.
 */

{
public:

	vec_ZZ_p hash_vector;
	vec_ZZ_p extra_hashes; //For verifying interpolation
	vec_ZZ_p valids; // validation points for the rational function - deviation from theory, as these should be randomly chosen for each synchronization (?)
	long setsize;

	Sketch() {
		setsize = 0;
	} //Default Constructor

	Sketch(vec_ZZ_p& p_hashes, vec_ZZ_p& p_extra_hashes, vec_ZZ_p& p_valids, long p_setsize);

	Sketch(const Sketch& other) { operator=(other);	} //Default Constructor

	Sketch& operator=(const Sketch& other)
	{
		if (this == &other)
			return *this;

		this->hash_vector = other.hash_vector;
		this->extra_hashes = other.extra_hashes;
		this->valids = other.valids;
		this->setsize = other.setsize;
	}

	Sketch(vec_ZZ_p& samples, int mbar, int bitsize);
	/* Constructor for Sketch
	 R: Subset in vector space whose hashes have to be stored, samplepoints, number of subpartitions
	 E: Creates Sketch
	 M:
	 */

	void getEvaluationValues(ostream& out_1, ostream& out_2, ostream& out_3,  long& size1, long&
	size2, long& size3, long& p_setSize );

	void reset();

	~Sketch() {
		valids.kill();
		hash_vector.kill();
		extra_hashes.kill();
	}
};
#endif
