#include <time.h>
#include <math.h>
//#include <strstream>
#include "prob_recon_support.h"
#include "constants.h"
#include "reconciled_set.h"
//#include <sys/time.h>
//#include <sys/resource.h>
//#include <sys/sysinfo.h>

// external variables
extern long SetRecon::time_comput;
extern long SetRecon::time_setup;
extern long SetRecon::time_add;
extern long SetRecon::num_add;
extern long SetRecon::time_degree;

// IMPLEMENTATION of Sketch class functions
Sketch::Sketch(vec_ZZ_p& samples, int mbar, int bitsize)
/* Constructor for Sketch
 R: Subset in vector space whose hashes have to be stored, samples, number of subpartitions
 E: Creates Sketch
 M: valids - makes it contain the validation points for this Sketch
 */
{
	// subset_pp.length();

	valids = compute_validationpoints(mbar, 1, power(ZZ_two, bitsize) + mbar
			/ 2 + 2, ZZ_p::modulus() - (mbar / 2) + 1, ZZ_zero); // start item serves as a seed for the random numbers

	clock_t start = clock();
	//	hash_vector = calc_poly(subset_pp, samples);
	//	extra_hashes = calc_poly(subset_pp, valids); //Extra sample points evaluation for probabilistic

	hash_vector.SetLength(samples.length());
	extra_hashes.SetLength(valids.length());

	reset();

	SetRecon::time_add += clock() - start;
	// SetRecon::num_add += subset_pp.length();
	//	cout << "\nNode type:" <<Sketch_type;
	//	cout << ", Node setsize:"<<setsize;
}

Sketch::Sketch(vec_ZZ_p& p_hashes, vec_ZZ_p& p_extra_hashes, vec_ZZ_p& p_valids, long p_setsize)
{
    hash_vector = p_hashes;
    extra_hashes = p_extra_hashes;
    valids = p_valids;
    setsize = p_setsize;
}

void Sketch::reset() {
	ZZ_p temp;

	setsize = 0;


	conv(temp, to_ZZ("1"));
	for (int i = 0; i < hash_vector.length(); i++)
		hash_vector[i] = temp;

	for (int i = 0; i < extra_hashes.length(); i++)
		extra_hashes[i] = temp;

}

ZZ SetRecon::getField(int bit_s, int m_bar) {
	return NextPrime(3 * power(ZZ_two, bit_s - 1) + m_bar);
}

void SetRecon::setField(int bit_s, int m_bar) {
	ZZ_p::init(getField(bit_s, m_bar));
}

SetRecon::SetRecon(int the_bitsize, int the_mbar) {
	int ii;
	clock_t start = clock();

	bitsize = the_bitsize;
	mbar = the_mbar;
	if (!(mbar % 2))
		mbar++; // Make this odd to optimize matrix back substitution

	// set dependent variables
	fieldsize = getField(bitsize, mbar);
	fieldbits = (int) ceil(log(fieldsize) / std::log(2.0f));
	dataStart = to_ZZ_p(mbar / 2 + 1);
	ZZ_p::init(fieldsize);

	samples = compute_samplepoints(mbar, to_ZZ_p(fieldsize - 1));

	powSamples = NULL;
	// compute and store powers of the sample points
	powSamples = new vec_ZZ_p[samples.length()];

	for (ii = 0; ii < samples.length(); ii++) // for various samples
	{
	    powSamples[ii].SetLength(mbar + 1);
	    for (int pow = 0; pow <= mbar; pow++)
		// and various powers
		// append(powSamples[ii],power(samples[ii],pow));  // compute the corresponding power
		powSamples[ii][pow] = power(samples[ii], pow);
	}

	// adjust the initialSet elements
	// for (ii = 0; ii < initialSet.length(); ii++)
	//	initialSet[ii] += dataStart;
	vec_ZZ_p initialSet;

	m_poly = new Sketch(samples, mbar, bitsize);

	time_setup += clock() - start;
}


SetRecon::SetRecon(const SetRecon& other) {
	this->powSamples = NULL;
	this->m_poly = NULL;

	operator=(other);
}

void Sketch::getEvaluationValues(ostream& out_1, ostream& out_2, ostream& out_3,  long& size1, long&
	size2, long& size3, long& p_setSize )
{
	for (int i = 0; i < hash_vector.length(); i++)
	    out_1 << hash_vector[i] << "\n";

	for (int j = 0; j < extra_hashes.length(); j++)
	    out_2 << extra_hashes[j] << "\n";

	for (int j = 0; j < valids.length(); j++)
	    out_3 << valids[j] << "\n";

	size1 = hash_vector.length();
	size2 = extra_hashes.length();
	size3 = valids.length();
	p_setSize = setsize;
}

unsigned int SetRecon::getEvalSize() {
	return (fieldbits / 8) * (m_poly->hash_vector.length()
			+ m_poly->extra_hashes.length());
}

void SetRecon::addPoly(Sketch* poly_to_add) {
	long i;
	long mbar = samples.length();

	if (m_poly == NULL || poly_to_add == NULL || m_poly->extra_hashes.length()
			!= poly_to_add->extra_hashes.length()
			|| m_poly->hash_vector.length()
					!= poly_to_add->hash_vector.length()) {
		cout << "Poly cannot be added!";
		return;
	}

	m_poly->setsize += poly_to_add->setsize;

	for (i = 0; i < mbar; i++)
		m_poly->hash_vector[i] *= poly_to_add->hash_vector[i]; //Updating sample point evaluations

	for (i = 0; i < m_poly->valids.length(); i++)
		m_poly->extra_hashes[i] *= poly_to_add->extra_hashes[i]; //Updating validation point evaluations
}

void SetRecon::resetPoly() {
	if (m_poly != NULL)
		m_poly->reset();
}

SetRecon& SetRecon::operator=(const SetRecon& other) {
	if (this == &other)
		return *this;

	this->bitsize = other.bitsize; // number of bits per vector
	this->mbar = other.mbar; // maximum number of differences to reconcile (i.e. recovery bound)
	this->fieldsize = other.fieldsize; // size of the underlying filed; a prime number in this case
	this->fieldbits = other.fieldbits;
	this->dataStart = other.dataStart;
	this->samples = other.samples;

	if (this->powSamples != NULL)
		delete powSamples;

	if (this->m_poly != NULL)
		delete m_poly;

	this->powSamples = NULL;
	this->m_poly = NULL;

	if (other.powSamples != NULL) {
		this->powSamples = new vec_ZZ_p[other.samples.length()];

		for (int i = 0; i < samples.length(); i++)
			this->powSamples[i] = other.powSamples[i];
	}

	if (other.m_poly != NULL) {
		this->m_poly = new Sketch();

		*(this->m_poly) = *(other.m_poly);
	}

	return *this;
}

SetRecon::~SetRecon() {
	FreePow();
	if (m_poly != NULL)
		delete m_poly;
	samples.kill();
}

void SetRecon::FreePow() {
	if (powSamples == NULL)
		return;

	for (int ii = 0; ii < samples.length(); ii++)
		powSamples[ii].kill();
	delete[] powSamples;

	powSamples = NULL;
}

int SetRecon::Reconcile(Sketch* p_otherPoly, vec_ZZ_p& diffA, vec_ZZ_p& diffB) {
	clock_t start = clock();

	unsigned int polyBDegree, polyADegree;

	if (m_poly == NULL || p_otherPoly == NULL) {
		cout << "Null pointers!";
		return 0;
	}

	if (!reconcile_aux(this, m_poly, p_otherPoly, polyADegree, polyBDegree, diffA, diffB, false)) 
	//Reconcile failed condition
	{
	///	cout<< "Error: Recon failed!" << endl;
		return -1;
	} //else
//	cout << "Reconciliation successful!" << endl;

		// fix up the data
	for (int i = 0; i < diffA.length(); i++)
	    diffA[i] -= dataStart; 

	for (int i = 0; i < diffB.length(); i++)
	    diffB[i] -= dataStart; 

	time_comput += clock() - start;

	int res = polyBDegree + polyADegree - 2;

	return res > 0 ? res : 0; 
}

int SetRecon::Reconcile(Sketch* p_otherPoly, unsigned int& diffASize, 
	unsigned int& diffBSize) {
	clock_t start = clock();
	vec_ZZ_p diffA, diffB;

	if (m_poly == NULL || p_otherPoly == NULL) {
		cout << "Null pointers!";
		return 0;
	}

	if (!reconcile_aux(this, m_poly, p_otherPoly, diffASize, diffBSize, diffA, diffB, true)) 
	//Reconcile failed condition
	{
		//cout<< "Error: Recon failed!" << endl;
		return -1;
	}// else
	//cout << "Reconciliation successful!" << endl;

	/*	// fix up the data
	 for (int ii = 0; ii < diff.length(); ii++)
	 diff[ii] -= dataStart; */

	time_comput += clock() - start;
	// return (diff);

	 if (diffASize > 0)
	     diffASize--;

	 if (diffBSize > 0)
	     diffBSize--;

	return diffASize + diffBSize;
}


Sketch* SetRecon::getPoly() {
	return m_poly;
}

/*
 bool SetRecon::test(int the_bitsize, int setCommon, int setDiff,
 int the_mbar, bool confirm)
 {
 int ii,jj;
 clock_t start = clock(); // keep track of time

 // initialize
 SetRecon::setField(the_bitsize,the_mbar);
 SetSeed(to_ZZ(time(NULL)));
 ZZ_p randBits;

 // generate the common elements
 vec_ZZ_p commonV;
 for (ii=0; ii<setCommon; ii++) { // only add unique elements
 do {
 randBits = to_ZZ_p(RandomBits_ZZ(the_bitsize));
 } while(elementOf(randBits,commonV));
 append(commonV,randBits);
 }

 // generate sets for A and B
 // randomly allocate differences between A and B
 int Adiff, Bdiff;
 Adiff = to_int(RandomBnd(to_ZZ(setDiff)));
 Bdiff = setDiff - Adiff;
 cout << "Difference: " << Bdiff << endl;

 vec_ZZ_p setA(commonV),setB(commonV);

 for (ii=0; ii<Adiff; ii++) {
 ZZ_p randBits = to_ZZ_p(RandomBits_ZZ(the_bitsize));
 do {
 randBits = to_ZZ_p(RandomBits_ZZ(the_bitsize));
 } while(elementOf(randBits,setA) || elementOf(randBits,commonV));
 append(setA,randBits);
 }

 for (ii=0; ii<Bdiff; ii++) {
 ZZ_p randBits = to_ZZ_p(RandomBits_ZZ(the_bitsize));
 do {
 randBits = to_ZZ_p(RandomBits_ZZ(the_bitsize));
 } while(elementOf(randBits,setB) || elementOf(randBits,commonV));
 append(setB,randBits);
 }

 time_setup += clock() - start;
 // reconcile the sets
 //cout << whoami << " commonV: " << commonV << endl;
 //cout << whoami << " setA: " << setA << endl << endl;

 GetMemInfo("Before Setup");

 SetRecon hostA( the_bitsize, setA, the_mbar );
 SetRecon hostB( the_bitsize, setB, the_mbar, true );

 setA.kill();
 setB.kill();
 commonV.kill();

 GetMemInfo("Before Reconciliation");

 vec_ZZ_p diff_vector = hostA.Reconcile(hostB.getPoly());
 // vec_ZZ_p diff2_vector = hostB.Reconcile(hostA.getTree());

 if (confirm) {
 vec_ZZ_p real_diff = difference_set(setA,setB);
 //append(real_diff,difference_set(setB,setA));

 // real_diff = diff_vector iff real_diff - diff_vector = diff_vector - real_diff = 0
 vec_ZZ_p check = difference_set(real_diff,diff_vector);
 if (check.length()!=0) {
 cout << "Set A:       " << setA << endl
 << "Set B:       " << setB << endl
 << "Differences: " << diff_vector << endl
 << "Real diff:   " << real_diff << endl
 << "Error:       " << check << endl;
 return(false);
 }
 check = difference_set(diff_vector,real_diff);
 if (check.length()!=0) {
 cout << "Set A:       " << setA << endl
 << "Set B:       " << setB << endl
 << "Differences: " << diff_vector << endl
 << "Real diff:   " << real_diff << endl
 << "Error:       " << check << endl;
 return(false);
 }
 }
 // else {
 // just print out the results
 cout << "Set A size:       " << setA.length() << endl
 << "Set B size:       " << setB.length() << endl
 << "Number of differences: " << diff_vector.length() << endl;
 //	    << "Differences: " << diff_vector << endl;
 //  }
 return(true);  // all confirmations passed
 }
 */

void SetRecon::add_element(ZZ_p& el) {
	clock_t start = clock();
	long ii;
	long mbar = samples.length();

	if (m_poly == NULL) {
		cout << "Null pointer in add_element!";
		return;
	}

	el += dataStart; // element has to be adjusted as in SetRecon::SetRecon

	m_poly->setsize++;

	for (ii = 0; ii < mbar; ii++)
		m_poly->hash_vector[ii] *= (samples[ii] - el); //Updating sample point evaluations

	for (ii = 0; ii < m_poly->valids.length(); ii++)
		m_poly->extra_hashes[ii] *= (m_poly->valids[ii] - el); //Updating validation point evaluations

	time_add += clock() - start;
	num_add++;
}

// void SetRecon::del_element(ZZ_p el) {
// 	delete_from_tree(m_poly,pp,el+dataStart);
// }

void SetRecon::construct_RHS_constant_vector(vec_ZZ_p& RF, long& rank,
		long& d, vec_ZZ_p& rtnvalue)
//Constructs the rhs for the Vandermonde matrix - the bmatrix . RF is the Rational Function Matrix. Rank is the rank of the LHS Matrix

{
	long ii;
	long mabar = calc_mabar(mbar, d);
	long mbbar = calc_mbbar(mbar, d);

	//long mbbar=mbar-mabar;
	rtnvalue.SetLength(rank);
	for (ii = 0; ii < rank; ii++)
		rtnvalue[ii] = RF[ii] * powSamples[ii][mbbar] - powSamples[ii][mabar];
	// i.e.  RF[ii]*power(samplepoints[ii],mbbar)-power(samplepoints[ii],mabar);
}

void SetRecon::construct_vandermonde_matrix(const vec_ZZ_p& RF,
		const vec_ZZ_p& bmatrix, const long d, mat_ZZ_p& mat_rtn)
// This function constructs the Vandemonde Matrix.
//RF is the Rational function vector (all the fi s) bmatrix is the RHS Matrix of size m bar.

{
	//cout <<"\nmbar = "<<mbar<<"\nd = "<<d;
	long ii, jj; //Counters

	long mabar = calc_mabar(mbar, d);
	long mbbar = calc_mbbar(mbar, d);
	//long mbbar=mbar-mabar;
	//   cout <<"\nmabar = "<<mabar;
	//   cout <<"\nmbbar = "<<mbbar;

	vec_vec_ZZ_p returnmatrix;
	returnmatrix.SetLength(mbar);

	for (ii = 0; ii < mbar; ii++) {
		returnmatrix[ii].SetLength(mabar + mbbar + 1); //+1 is for the B matrix

		for (jj = 0; jj < mabar; jj++)
			returnmatrix[ii][jj] = powSamples[ii][mabar - jj - 1];
		for (jj = 0; jj < mbbar; jj++)
			returnmatrix[ii][jj + mabar] = -RF[ii] * powSamples[ii][mbbar - jj
					- 1];
		returnmatrix[ii][mabar + mbbar] = bmatrix[ii]; //Augumented matrix
	}

	//  //Adding alternate rows (starting from row 1 to mbar-1; leaving out row 0)
	//   for(ii=1;ii<mbar-1;ii=ii+2)
	//     for(long jj=0;jj<=mabar+mbbar;jj++)
	//       returnmatrix[ii][jj]+=returnmatrix[ii+1][jj];

	//   //Dividing each alternate row by 2
	//   for(ii=1;ii<mbar;ii=ii+2)
	//     for(jj=0;jj<=mabar+mbbar;jj++)
	//       returnmatrix[ii][jj]/=2;

	//   //Subtracting alternate rows (Leaving Row 0 alone) row 2 = row 2 - row 1 etc
	//   for(ii=2;ii<mbar;ii=ii+2)
	//     for(jj=0;jj<=mabar+mbbar;jj++)
	//       returnmatrix[ii][jj]-=returnmatrix[ii-1][jj];

	mat_rtn.SetDims(mbar, mabar + mbbar + 1);//+1 is for the B matrix
	MakeMatrix(mat_rtn, returnmatrix);
}
